from torchelper.events.event import Event
from torchelper.events.receiver import Receiver
from threading import Lock

lock = Lock()
__event_map__ = {}

class EventCenter:
    def __init__(self):
        pass
    
    @staticmethod
    def register(evt_type, rcver:Receiver):
        '''
        evt_type: str|list
        '''
        lock.acquire()
        global __event_map__
        if isinstance(evt_type, str):
            evt_type = [evt_type]
        for evt in evt_type:
            rcvers = __event_map__.get(evt, set())
            rcvers.add(rcver)
            __event_map__[evt] = rcvers
        lock.release()

 

    @staticmethod
    def __on_event(event:Event):
        lock.acquire()
        global __event_map__
        rcvers = __event_map__.get(event.evt_type, set())
        for rcver in rcvers:
            rcver.on_event(event)
        lock.release()

    @staticmethod
    def broadcast(event:Event):
        EventCenter.__on_event(event)




from .dist_util import master_only
import functools
import time
from termcolor import colored
 
__log_file=None


def __my_log(level:int, *args):
    log = [str(v) for v in args]
    now = time.strftime("[%Y-%m-%d %H:%M:%S] ", time.localtime())
    log = now + ' '.join(log)
    if level==0:
        print(log)
    elif level == 1:
        print(colored(log, 'yellow'))
    else:
        print(colored(log, 'red'))
    if __log_file is not None:
        __log_file.write(log+"\n")

@master_only
def debug(*msg):
    __my_log(0, *msg)

@master_only
def log(*msg):
    __my_log(0, *msg)

@master_only
def warn(*msg):
    __my_log(1, *msg)

@master_only
def error(*msg):
    __my_log(2, *msg)

  
@master_only
def init_log(log_file_path=None):

    global __log_file
    __log_file = open(log_file_path, "w+",encoding="utf-8")
 



class Callback():
    def __init__(self):
        pass

    def on_begin_train(self, epoch:int):
        pass

    def on_end_train(self):
        pass


    def on_begin_epoch(self, epoch:int):
        pass

    def on_end_epoch(self, epoch:int):
        pass


    def on_begin_step(self, epoch:int, step:int):
        pass

    def on_end_step(self, epoch:int, step:int):
        pass
    

    def on_begin_val(self, epoch:int):
        pass
    
    def on_val(self, epoch:int, data):
        pass

    def on_end_val(self, epoch:int, metric_dict:dict):
        pass

  



from torchelper.callbacks.callback import Callback


class Trainable(Callback):
    '''可训练的抽象类
    '''
    def __init__(self):
        pass

 
    def get_loss(self, name:str):
        '''返回当前model loss，如果返回None，则不做梯度更新
        '''
        return None

    def set_train(self):
        pass

    def set_eval(self):
        pass
 
    def on_begin_forward(self, data, epoch, step):
        None

    def on_forward(self,  epoch, step, inp):
        return None
    
    def on_end_forward(self,  epoch, step):
        return None
    
    def on_begin_backward(self,  epoch, step):
        return None

    def on_backward(self,  epoch, step):
        return None
    
    def on_end_backward(self, epoch, step):
        return None


class Event:
    
    Scalar="evt_scalar"
    Voice="evt_voice"
    Pic="evt_pic"
    Level_step = 0
    Level_epoch = 1
    Level_val = 2
 

    def __init__(self, evt_type:str, data:dict, level:int=0, extra:dict={}):
        self.evt_type = evt_type
        self.data = data
        self.extra = extra
        self.level=level

    def wrap_extra(self, key, val):
        self.extra = {
            key:val
        }


from torchelper.utils.dist_util import get_rank
from torchelper.events.receiver import Receiver
from torchelper.events.event import Event
from tqdm import tqdm


class PBar(Receiver):
    def __init__(self, iter):
        rank = get_rank()
        self.is_master = rank == 0
        if self.is_master:
            self.bar = tqdm(iter)
        else:
            self.bar = iter

    def __iter__(self):
        return self.bar.__iter__()
    
    def on_event(self, event: Event):
        if not self.is_master:
            return
        if event.evt_type == Event.Scalar:
            data = event.data
            msg = []
            for k, v in data.items():
                msg.append('%s:%.5f'%(k, v))
            self.bar.set_description(', '.join(msg))



import numpy as np
from torch.utils.tensorboard import SummaryWriter
from torchelper.events.receiver import Receiver
from torchelper.events.event import Event
from torchelper.utils import logger


class TensorBoardReceiver(Receiver):
    def __init__(self, tb_dir):
        self.tb_writer = SummaryWriter(tb_dir)

    def add_scalars(self, step, metrics):
        for k, v in metrics.items():
            if v is None:
                logger.warn(k+" is None ... ")
                continue
            self.tb_writer.add_scalar(k, v, step)
        
    def add_audio(self, step, audios):
        for k, v in audios.items():
            if v is None:
                logger.warn(k+" is None ... ")
                continue    
            self.tb_writer.add_audio(k, v, step, sample_rate=16000)

    def add_images(self, step, images):
        for k, v in images.items():
            if v is None:
                logger.warn(k+" is None ... ")
                continue
            self.tb_writer.add_image(k, v, step)
        

    def on_event(self, event: Event):
        data = event.data
        extra = event.extra
        type = event.evt_type
        step = extra.get('step',None)
        if step is None:
            return
        if type==Event.Voice:
            self.add_audio(step, data)
        elif type==Event.Scalar:
            self.add_scalars(step, data)
        elif type==Event.Pic:
            self.add_images(step, data)
        else:
            logger.error("The type", type, "is not supported currently")
       
       


import torch
import torch.nn as nn
from torchelper.utils.dist_util import get_rank
from torchelper.models.trainable import Trainable
from torch.cuda.amp import autocast, GradScaler

class BaseModel(nn.Module, Trainable):
    def __init__(self, cfg):
        super().__init__()
        self.builder = None
        self.__optimizer = None
        self.cfg = cfg
        self.is_half = cfg.get('amp', False)
        self.grad_scaler = None
        if self.is_half:
            self.grad_scaler = GradScaler()
        self.init_lr = cfg['init_lr']
        self.callbacks = []
        self.gpu_id = get_rank()
        if self.gpu_id>=0:
            self.device = torch.device('cuda:'+str(self.gpu_id))
        else:
            self.device = torch.device('cpu')


    def get_builder(self):
        return self.builder
        
    def remap_weights_name(self, weights):
        return weights

    def add_cb(self, cb):
        self.callbacks.append(cb)
    
    def perform_cb(self, event:str, *arg, **args):
        for cb in self.callbacks:
            evt_func = getattr(cb, event)
            evt_func(*arg, **args)

    def get_train_params(self):
        '''返回当前model训练参数，如果返回None，则默认为所有训练参数
        '''
        return None
    
    def get_loss(self, name:str):
        '''返回当前model loss，如果返回None，则不做梯度更新
        '''
        return None

    def get_optimizer(self):
        if self.__optimizer is None:
            params = self.get_train_params()
            if params is None:
                params = list(self.parameters())
            if len(params)>0:
                self.__optimizer = torch.optim.Adam(params, lr=self.init_lr, betas=(0.5, 0.999))
        return self.__optimizer

    def set_train(self):
        self.train()

    def set_eval(self):
        self.eval()
    # -------------------- forward -------------------- #
    def on_forward(self, epoch, step, inp):
        return self.forward(inp)
    
    # -------------------- backward -------------------- #
    def on_backward(self, epoch, step):
        if self.is_half:
            self.fp16_backward()
        else:
            self.fp32_backward()

    def fp16_backward(self):
        '''混合精度训练一个step执行反向
        '''
        with autocast():
            loss = self.get_loss()
        if loss is not None:
            optimizer = self.get_optimizer()
            if optimizer is None:
                return
            scaler = self.grad_scaler
            optimizer.zero_grad()
            # Scales loss. 为了梯度放大.
            scaler.scale(loss).backward()
            # scaler.step() 首先把梯度的值unscale回来.
            # 如果梯度的值不是 infs 或者 NaNs, 那么调用optimizer.step()来更新权重,
            # 否则，忽略step调用，从而保证权重不更新（不被破坏）
            scaler.step(optimizer)
            # 准备着，看是否要增大scaler
            scaler.update()

    def fp32_backward(self):
        '''一个step执行反向
        '''
        loss = self.get_loss()
        if loss is not None:
            optimizer = self.get_optimizer()
            if optimizer is None:
                return
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
    



from torchelper.models.trainable import Trainable
from torchelper.models.base_model import BaseModel
import torch.nn as nn
from torch.nn.parallel import DistributedDataParallel

class GroupModel(Trainable):
    def __init__(self, cfg):
        self.models = {}
        self.cfg = cfg
    
    def add_model(self, name:str, model:BaseModel):
        def get_sub_loss():
            return self.get_loss(name)
        model.get_loss = get_sub_loss
        self.models[name] = self.model_to_device(model)

    def on_backward(self, epoch, step):
        for _, model in self.models.items():
            model.module.on_backward(epoch, step)

    def model_to_device(self, net:nn.Module, find_unused_parameters=False):
        """Model to device. It also warps models with DistributedDataParallel.
        :param net: nn.Module
        """
        net = net.cuda(self.gpu_id)
        net = DistributedDataParallel(
            net, device_ids=[self.gpu_id], find_unused_parameters=find_unused_parameters)
        return net

    def forward_sub_model(self, name, data):
        return self.models[name](data)

    def set_train(self):
        for _, model in self.models.items():
            model.train()

    def set_eval(self):
        for _, model in self.models.items():
            model.eval()

    def perform_cb(self, event:str, *arg, **args):
        evt_func = getattr(self, event)
        evt_func(*arg, **args)
        for _, model in self.models.items():
            model.module.perform_cb(event, *arg, **args)

    
    


import os
import random
from torchelper.utils.dist_util import get_rank
from torchelper.models.model_builder import ModelBuilder
from torchelper.events.event import Event
from torchelper.events.event_center import EventCenter
from torchelper.utils import  logger
from torchelper.events.tensorboard_receiver import  TensorBoardReceiver
from torchelper.events.wechat_receiver import WechatReceiver
from tqdm import tqdm
import torch 
import time

import torch.distributed as dist
import torch.multiprocessing as mp
from torchelper.utils.config import merge_cfg
from torchelper.utils.cls_utils import get_cls, new_cls, auto_new_cls
from torchelper.data.base_dataset import get_data_loader
import torch.backends.cudnn as cudnn
import subprocess
from torchelper.utils import logger
from torchelper.models.trainable import Trainable
from torchelper.events.event_center import EventCenter

from torchelper.events.pbar_receiver import PBar
from torchelper.events.event import Event
from torchelper.utils.global_data import set_global
import pickle


def check_close_port(port):
    result = subprocess.run(['lsof', '-i:'+str(port)], stdout=subprocess.PIPE)
    out = result.stdout.decode('utf-8')

    lines = out.split('\n')
    if len(lines)<=1:
        return
    logger.warn(out)
    lines = lines[1:]
    for line in lines:
        arr = [s for s in line.split(' ') if len(s)>0]
        if len(arr)<2:
            continue
        pid = int(arr[1])
        os.system('kill '+str(pid))
        logger.warn("kill", pid)

def check_close_gpu_ids(gpu_ids):
    if not isinstance(gpu_ids, list):
        gpu_ids = [gpu_ids]
    logger.warn('kill:', gpu_ids)
    result = subprocess.run(['nvidia-smi'], stdout=subprocess.PIPE)
    out = result.stdout.decode('utf-8')

    lines = out.split('\n')
    if len(lines)<=1:
        return

    start_flag = False
    for line in lines:
        if not start_flag:
            if line.startswith('|=') and not '+' in line:
                start_flag = True
        else:
            line = ' '.join(line.split())
            arr = line.split(' ')
            if len(arr)<8:
                continue
            gpu_id = int(arr[1])
            pid = int(arr[4])
            for gid in gpu_ids:
                if gpu_id == gid and arr[6].endswith('/python'):
                    os.system('kill '+str(pid))
                    logger.warn('kill', arr[6])
      

def get_port(def_port):
    result = subprocess.run(['lsof', '-i:'+str(def_port)], stdout=subprocess.PIPE)
    out = result.stdout.decode('utf-8')

    lines = out.split('\n')
    if len(lines)<=1:
        return def_port
    return get_port(def_port+1)


def validate(model:Trainable, epoch:int, val_dataset,   gpu_num:int):
    '''执行验证集
    :param net: ModelBuilder子类实例
    :param epoch: int, 当前epoch
    :param val_dataset: 验证集
    :return: 返回描述
    :raises ValueError: 描述抛出异常场景
    '''
    model.set_eval()
    if get_rank()==0:
        model.perform_cb('on_begin_val', epoch=epoch)
    val_data:dict = {}
    pbar = PBar(val_dataset)
    for data in pbar:
        res = model.on_val(epoch, data)
        if res is None:
            continue
        for key, val in res.items():
            val_data[key] = val + val_data.get(key, 0)

    avg_val_dict = {}
    if val_data is not None:
        val_arr = []
        for key, val in val_data.items():
            avg_val_dict[key] = val_data[key] * 1.0 / len(val_dataset)

    data = pickle.dumps(avg_val_dict)
    with open(str(get_rank()), "wb+") as f:
        f.write(data)
    torch.distributed.barrier() 
    if get_rank()==0 and val_data is not None:
        names = avg_val_dict.keys()
        val_arr = []
        rs = {}
        for key in names:
            avg_v = 0
            for i in range(gpu_num):
                with open(str(i), "rb") as f:
                    data = f.read()
                avg_dict = pickle.loads(data)
                avg_v += avg_dict[key]
            avg_v = avg_v / gpu_num
            rs[key] = avg_v
            val_arr.append(key + ":" + str(avg_v))
        line = 'epoch: ' + str(epoch) + ', '  + ', '.join(val_arr)
        model.perform_cb('on_end_val', epoch=epoch, metric_dict=rs)
        logger.debug(line)
    torch.distributed.barrier() 

def init_master(cfg):
    logger.init_log(os.path.join(cfg['ckpt_dir'], "log.log"))
    wc_id = cfg.get("wechat_key", None)
    EventCenter.register([Event.Scalar, Event.Voice, Event.Pic], TensorBoardReceiver(cfg['ckpt_dir']))
    if wc_id is not None:
        EventCenter.register(Event.Scalar, WechatReceiver(wc_id))

def train(cfg):
   
    gpu_id = get_rank()
    if gpu_id == 0:
        init_master(cfg)
    train_data_cfg = merge_cfg(cfg['dataset']['train'], cfg)
    val_data_cfg =  merge_cfg(cfg['dataset']['val'], cfg)
    train_dataset = auto_new_cls(train_data_cfg)
    val_dataset = auto_new_cls(val_data_cfg)
    gpu_count = len(cfg['ori_gpu_ids'])
 
    train_dataloader = get_data_loader(cfg['batch_per_gpu'], train_dataset, num_workers=0, dist=True)
    val_dataloader = get_data_loader(cfg['batch_per_gpu'], val_dataset, dist=True)

    model:Trainable = get_cls(cfg['model'])(cfg)
 
    # builder.set_dataset(train_dataset)
    dataset_size = len(train_dataloader)
     
    step_per_epoch_per_gpu = dataset_size 
    set_global("step_per_epoch_per_gpu", step_per_epoch_per_gpu)
    set_global('gpu_count', gpu_count)
    model.perform_cb('on_begin_train', epoch=cfg['start_epoch'])
  
    for epoch in range(cfg['start_epoch'], cfg['total_epoch']):
        # validate(model, epoch, val_dataloader,  gpu_count)
        model.set_train()
        pbar = PBar(train_dataloader)
        enum_data = enumerate(pbar)
        # if gpu_id==0:
        #     pbar = tqdm(train_dataloader)
        #     enum_data = enumerate(pbar)
        # else:
        #     pbar = None
        #     enum_data =  enumerate(train_dataloader)
        EventCenter.register(Event.Scalar, pbar)
        # on begin epoch
        model.perform_cb('on_begin_epoch', epoch=epoch)
        for i, data in enum_data:
            model.on_begin_forward(data, epoch, i)
            model.perform_cb('on_begin_step', epoch=epoch, step=i)
            model.on_forward(epoch, i, data)
            model.on_end_forward(epoch, i)
            model.on_begin_backward(epoch, i)
            # 计算loss
            model.on_backward(epoch, i)
            # 多卡同步
            torch.distributed.barrier()
            model.on_end_backward(epoch, i)
            model.perform_cb('on_end_step', epoch=epoch, step=i)
            
        model.perform_cb('on_end_epoch', epoch=epoch)
        validate(model, epoch, val_dataloader, gpu_count)
    model.perform_cb('on_end_train')


def train_worker(gpu_id, nprocs, cfg, port):
    '''独立进程运行
    '''
    os.environ['NCCL_BLOCKING_WAIT']="1"
    os.environ['NCCL_ASYNC_ERROR_HANDLING']='1'
    random.seed(0)
    torch.manual_seed(0)
    cudnn.deterministic = True
    # 提升速度，主要对input shape是固定时有效，如果是动态的，耗时反而慢
    torch.backends.cudnn.benchmark = True
    dist.init_process_group(backend='nccl',
                            init_method='tcp://127.0.0.1:'+str(port),
                            world_size=len(cfg['gpu_ids']),
                            rank=gpu_id)

    torch.cuda.set_device(gpu_id)
    # 按batch分割给各个GPU
    # cfg['batch_size'] = int(cfg['batch_size'] / nprocs)
    train(cfg)

def train_main(cfg):
    check_close_gpu_ids(cfg['ori_gpu_ids'])
    # check_close_port(cfg['port'])
    gpu_nums = len(cfg['gpu_ids'])
    # if gpu_nums>1:
    port = get_port(cfg['port'])
    logger.debug('init port:', port)
    mp.spawn(train_worker, nprocs=gpu_nums, args=(gpu_nums, cfg, port ))
 
    # else:
    #     train(cfg['gpu_ids'][0], cfg, False)



from .models.model_builder import ModelBuilder
from .callbacks.callback import Callback
from .callbacks.ckpt_callback import CkptCallback
from .events.wechat_receiver import WechatReceiver
from .callbacks.reduce_lr_on_plateau import ReduceLROnPlateau
from .models.base_model import BaseModel
from .utils.config import init_cfg, load_cfg, merge_cfg
from .train import train_main
from .metrics import measure
from .data import *
from .models.lr_scheduler import LinearDownLR
from .utils.dist_util import master_only, get_rank
from .metrics  import *
from .utils.cls_utils import new_cls 
from .utils import logger
from .events.event import Event
from .events.event_center import EventCenter
from .events.receiver import Receiver
from .events.pbar_receiver import PBar
from .events.tensorboard_receiver import TensorBoardReceiver
from .models.trainable import Trainable
from .models.group_model import GroupModel
from .utils.global_data import set_global, get_global

name = "torchelper"